"""
Le code wsgi renvoie un fichier sous la forme d'un cgi.FieldStorage


Cas 1 : Dans le cas d'un formulaire classique: 
    
    le widget deform le deserialize en dict et le stocke dans un tempstore 
    pour le retrouver en cas d'erreur de validation

Cas 2 : Dans le cas de l'api rest, pas de deform donc on a le fichier sous la forme de cgi.FieldStorage

Niveau Wsgi et Rest api :

    {
        'upload': <FieldStorage>
    }
FieldStorage a les propriétés suivantes
    file : le fichier
    filename : le nom du fichier (à formatter dans le cas d'IE)
    type : le mimetype 
    length : la taille du fichier

sorti de deform il a les attributs suivant

    {
        'upload': {
            filename: nom du fichier
            mimetype: mimetype
            size: taille
            fp: pointer to file
            uid: uid
        }
    }

    
Le modèle File attend les données suivantes lors de l'initialisation

name : str  <- filename
description : str <- description ou filename
mimetype : str <- type
size : int
file_type_id : int

Puis dans un deuxième temps :

data : binary blob ou FieldStorage
"""
import logging

from cgi import FieldStorage
from typing import Optional

from endi.forms import merge_session_with_post
from endi.forms.files import get_file_upload_schema
from endi.models.files import File
from endi.models.indicators import SaleFileRequirement
from endi.models.node import Node

from endi.utils.controller import BaseAddEditController
from endi.events.files import FileAdded, FileDeleted, FileUpdated
from endi.views.files.routes import FILE_ITEM

logger = logging.getLogger(__name__)


def ensure_filename(name: str) -> str:
    """Handle the case when the browser (IE) returns an absolute path"""
    return name[name.rfind("\\") + 1 :].strip()


class FileController(BaseAddEditController):
    """
    Controller used to manipulate file objects

        - build a schema
        - add a file to the database
        - edit an existing file
        - delete an existing file

    >>> controller = FileController(self.request, self.context)
    >>> schema = controller.get_schema()
    >>> data = schema.deserialize(submitted_data)
    >>> controller.add(data, parent_node)
    """

    def __init__(self, request, edit=False, factory=File):
        super().__init__(request, edit)
        self.factory = factory

    def get_schema(self, submitted: Optional[dict] = None):
        return get_file_upload_schema()

    def file_to_appstruct(self, file_object: File) -> dict:
        """
        Format a file_object as a dict matching the upload schema for deform.Form
        initialization
        """
        file_dict = {
            "name": file_object.name,
            "description": file_object.description,
            "file_type_id": file_object.file_type_id,
        }
        file_dict["upload"] = {
            "filename": file_dict["name"],
            "uid": str(file_object.id),
            "preview_url": self.request.route_path(
                FILE_ITEM, id=file_object.id, _query={"action": "download"}
            ),
        }
        return file_dict

    def set_upload_data(self, file_object, upload):
        if isinstance(upload, dict):
            file_object = self.set_file_data_from_dict(file_object, upload)
        elif isinstance(upload, FieldStorage):
            file_object = self.set_file_data_from_fieldstorage(file_object, upload)
        else:
            file_object = None
        return file_object

    def _get_size(self, file_pointer) -> int:
        file_pointer.seek(0)
        value = len(file_pointer.read())
        file_pointer.seek(0)
        return value

    def set_file_data_from_dict(self, file_object: File, upload: dict) -> File:
        """Set the uploaded data on a file when in dict format (coming from deform)"""
        if upload and "fp" in upload:
            file_object.name = upload["filename"]
            file_object.data = upload.get("fp", "")
            if "size" in upload:
                file_object.size = upload["size"]
            else:
                file_object.size = self._get_size(upload["fp"])
            file_object.mimetype = upload["mimetype"]

        return file_object

    def set_file_data_from_fieldstorage(
        self, file_object: File, upload: FieldStorage
    ) -> File:
        if upload is not None:
            file_object.name = ensure_filename(upload.filename)
            file_object.data = upload.file
            file_object.size = self._get_size(upload.file)
            file_object.mimetype = upload.type
        return file_object

    def _find_associated_indicator(
        self, default_parent: Node, attributes: dict
    ) -> Optional[SaleFileRequirement]:
        """
        Try to retrieve the indicator from the submitted data and concerning default_parent
        """
        indicator = None
        if "indicator_id" in attributes:
            indicator = SaleFileRequirement.get(attributes["indicator_id"])

        if indicator is None:
            # On tente de retrouver un requirement avec le même type de fichier
            if "file_type_id" in attributes and hasattr(
                default_parent, "get_file_requirements"
            ):
                file_type_id = attributes["file_type_id"]
                # On retrouve un file_requirement qui concerne le Node par défaut et qui a ce type
                indicators = default_parent.get_file_requirements(
                    file_type_id=file_type_id, scoped=False
                )

                if len(indicators) >= 1:
                    return indicators[0]

        return indicator

    def _get_parent(self, attributes: dict, default_parent: Node) -> Node:
        """
        Retrieve the parent the submitted file should be attached to
        """
        if "parent_id" in attributes:
            parent_id = attributes["parent_id"]
            node: Node = Node.get(parent_id)
            if node is not None and self.request.has_permission("add.file", node):
                default_parent = node

        indicator = self._find_associated_indicator(default_parent, attributes)

        if indicator is None or not self.request.has_permission(
            "add.file", indicator.node
        ):
            return default_parent

        return indicator.node

    def _add(self, attributes: dict, parent: Optional[Node] = None) -> File:
        """
        Add a file object

        **Attributes/Arguments**

        attributes

            Attributes passed through the schema validation

            {
                'come_from': 'http://example.com',
                'popup': 0,
                'upload': FieldStorage or {'fp', 'size' ...},
                'description': "Ma description",
                'file_type_id': 5
            }
            UPLOAD_DATA could be either a dict or an object (FieldStorage)
            (dict if coming from the deform widget)

        parent

            The parent object to which we attach the file
        """
        logger.debug("Adding a file")
        file_object = self.factory()
        upload = attributes.pop("upload", {})
        self.set_upload_data(file_object, upload)

        if file_object.name is None:
            raise Exception("Missing file data ?")

        if parent:
            default_parent = parent
        elif isinstance(self.context, Node):
            default_parent = self.context
        else:
            default_parent = None
        # Magic #3925 : déplacement auto des fichiers de ventes au niveau où ils sont
        # requis
        file_object.parent = self._get_parent(attributes, default_parent)
        # Merge other attributes
        merge_session_with_post(file_object, attributes)
        self.request.dbsession.add(file_object)
        self.request.dbsession.flush()

        self.request.registry.notify(FileAdded(self.request, file_object, attributes))
        return file_object

    def _edit(self, attributes) -> File:
        file_object: File = self.context
        upload = attributes.pop("upload", None)
        self.set_upload_data(file_object, upload)
        # Merge other attributes
        merge_session_with_post(file_object, attributes)

        # Magic #3925 : déplacement auto des fichiers de ventes au niveau où ils sont
        # requis
        parent = self._get_parent(attributes, file_object.parent)
        if parent != file_object.parent:
            file_object = self.move(attributes, parent)
        else:
            self.request.dbsession.merge(file_object)
            self.request.dbsession.flush()
            self.request.registry.notify(
                FileUpdated(self.request, file_object, attributes)
            )
        return file_object

    def save(self, attributes: dict, parent: Optional[Node] = None) -> File:
        if self.edit:
            return self._edit(attributes)
        else:
            return self._add(attributes, parent)

    def move(self, attributes: dict, parent: Node) -> File:
        file_object: File = self.context
        self.request.registry.notify(FileDeleted(self.request, file_object))

        file_object.parent = parent
        self.request.dbsession.merge(file_object)
        self.request.dbsession.flush()
        self.request.registry.notify(FileAdded(self.request, file_object, attributes))
        return file_object

    def delete(self):
        logger.info(
            "# {} deletes a {} with id {}".format(
                self.request.identity,
                str(self.context.__class__.__name__),
                self.context.id,
            )
        )
        self.request.registry.notify(FileDeleted(self.request, self.context))
        self.request.dbsession.delete(self.context)
